#ifndef MSP_GL_TEXT_H_
#define MSP_GL_TEXT_H_

#include "font.h"
#include "mesh.h"
#include "object.h"
#include "objectinstance.h"
#include "technique.h"

namespace Msp {
namespace GL {

/**
Creates an object consisting of the visual representation of a string.  If you
derive from Text to customize it, make sure you call Text::setup_render or
otherwise bind the appropriate texture.
*/
class Text: public ObjectInstance
{
public:
	enum HorizontalAlign
	{
		LEFT,
		CENTER,
		RIGHT
	};

	enum VerticalAlign
	{
		DESCENT,
		BASELINE,
		MIDLINE,
		ASCENT
	};

private:
	const Font &font;
	Mesh mesh;
	Object object;
	Tag texture_slot;
	float horz_align;
	float vert_offset;
	float width;

public:
	Text(const Font &, const Technique * = 0, Tag = Tag());

	const Mesh *get_mesh() const { return &mesh; }

	/** Sets technique to render with, replacing the given texture slot with
	the font texture.  If no texture slot is specified, heuristics are used to
	choose a suitable one. */
	void set_technique(const Technique *, Tag = Tag());

	const Technique *get_technique() const { return object.get_technique(); }

	/// Sets the string to be displayed.
	void set_text(const std::string &, StringCodec::Decoder &);

	template<typename C>
	void set_text(const std::string &t)
	{
		typename C::Decoder dec;
		set_text(t, dec);
	}

	void set_text(const std::string &t)
	{ set_text<StringCodec::Utf8>(t); }

	/// Clears the object's contents.
	void clear();

	/// Sets horizontal and vertical alignment with predefined anchors.
	void set_alignment(HorizontalAlign, VerticalAlign = BASELINE);

	/** Sets horizontal and vertical alignment.  0.0 means left or baseline,
	1.0 means right or top. */
	void set_alignment(float, float = 0.0f);

	float get_width() const { return width; }

	virtual void setup_render(Renderer &, Tag) const;

	DEPRECATED operator const Object &() const { return object; }
};

} // namespace GL
} // namespace Msp

#endif
